from Whittler.classes.MemoryCompressor import MaybeCompressedString
from Whittler.classes.RelevanceInterface import RelevanceInterface
from collections import OrderedDict
import hashlib
import re


class Result(dict, RelevanceInterface):

    FRIENDLY_NAME = NotImplemented
    
    ATTRIBUTES = []

    _init_run = False

    def __init__(self, resultdict=None):
        self.original_resultdict = resultdict
        self._cached_hash = 0
        self._cached_hash_valid = True
        self._frozen = False
        dict.__init__(self)
        self._relevant = True
        RelevanceInterface.__init__(self)

        if "whittler_filename" not in self.ATTRIBUTES:
            self.ATTRIBUTES.insert(0,"whittler_filename")
        if not resultdict is None:
            for k,v in resultdict.items():
                self[k] = v
        self._init_run = True

    def __setitem__(self, key, value):
        valtype = type(value)
        # if valtype is list:
        #     pass
        # elif valtype is dict:
        #     pass
        if not valtype is str:
            value = repr(value)
        
        # This check is just for pickling/unpickling objects... the pickle implementation first calls __setitem__ with each of the
        # key/value pairs in this dict subclass, then sets the variable values above - this causes AttributeErrors to pop up because
        # self.ATTRIBUTES, self._frozen, etc have not been initialized yet.
        if not self._init_run:
            self.__init__()
        if key not in self.ATTRIBUTES:
            raise Exception(f"Unrecognized key {key} in this result.")
        if self._frozen:
            raise Exception("Cannot modify this result after it has been added to the database!")
        if self.Config.REMOVE_ANSI_CONTROL_CHARACTERS:
            value = self.filter_ansi(value)
        self._cached_hash_valid = False
        dict.__setitem__(self, key, MaybeCompressedString(value))
    
    def __getitem__(self, key):
        return dict.__getitem__(self,key).value
    
    # def __reduce__(self):
    #     cls = self.__class__
    #     args = ()
    #     state = self.__dict__.copy()
    #     list_iterator = None
    #     dict_iterator = iter({key:self.Config.MemoryCompressor.decompress(value) for key,value in dict(self).items()}.items())
    #     return (cls,args,state,list_iterator,dict_iterator)

    def __hash__(self):
        if self._cached_hash_valid:
            return self._cached_hash
        new_hashval = 0
        for attr in sorted(self.ATTRIBUTES):
            if attr in self:
                # TODO: faster to hash the compressed values if memory compression enabled, would that break anything?
                new_hashval = int(hashlib.md5(bytes(str(new_hashval)+attr+self[attr],'utf-8')).hexdigest(),16)
        self._cached_hash = new_hashval
        self._cached_hash_valid = True
        return new_hashval
    
    def __eq__(self, other):
        if not isinstance(other, Result):
            return False
        return hash(self) == hash(other)
    
    def items(self):
        for k in dict.keys(self):
            yield (k,self[k])
    
    @staticmethod
    def give_result_dict_list(fname):
        raise NotImplementedError()
    
    @classmethod
    def _give_result_dict_list(cls, fname):
        ret = cls.give_result_dict_list(fname)
        for resultdict in ret:
            resultdict["whittler_filename"] = fname
        ret = [cls(result) for result in ret]
        _ = [hash(r) for r in ret] # cache the hashes (speeds things up during multiprocessing)
        return ret
    
    # https://stackoverflow.com/questions/14693701/how-can-i-remove-the-ansi-escape-sequences-from-a-string-in-python
    ansi_escape = re.compile(r'\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])')
    
    # Takes a unicode string, and filters out all 7-bit ANSI escape characters.
    def filter_ansi(self, s):
        return self.ansi_escape.sub('', s)
    
    # "Just read the f*cking string from the file already!"
    @staticmethod
    def force_read_file_to_string(fname):
        with open(fname, "rb") as f:
            b = f.read()
        b.replace(b"\x00",b"\\x00")
        b = b"".join([( bytes(bt) if (bt < 0x80 and bt != 0x00)
                        else bytes("\\x{:0>2}".format(hex(bt)[2:]),'ascii')) for bt in b])
        return b.decode("ascii")
    
    # This returns an elegant representation of the result. By default, it does not print the "diff"
    # component itself because it is usually very long and clutters up the output.
    def pretty_repr(self, objectview=None):
        if objectview is None:
            objectview = self.objectview
        s = ""
        def adds(content):
            nonlocal s
            s += str(content) + "\n"
        for k,v in self.items():
            if not objectview["SOLO"] is None:
                if k != objectview["SOLO"]:
                    continue
                adds(v)
                break
            elif not objectview["solo"] is None:
                if k != objectview["solo"]:
                    continue
            if not v.strip():
                continue
            adds(k)
            if k in objectview["quiet"] and k != objectview["solo"]:
                adds("... silenced ...")
                adds("")
                continue
            adds("\n".join("    "+line for line in v.splitlines()))
            adds("")
        return s
    
    def display(self):
        print(self.pretty_repr())
    
    def __repr__(self):
        return f"{self.__class__.__name__}({dict.__repr__(self)})"
    

    #######################################
    #  RelevanceInterface implementations
    #

    @property
    def relevant(self):
        return self._relevant
    
    def mark_irrelevant(self):
        self._relevant = False

    def mark_relevant(self):
        self._relevant = True
    
    def real_iter_values(self):
        return [self]
    

    #################################################
    #  NestedObjectPointerInterface implementations
    #
    
    def give_child_pointers(self):
        return OrderedDict({str(hash(self)):self.pointer_to_me})
    
    def size(self):
        return 1
    
    def all_result_objects(self):
        yield self

    def show_view(self, ct=0, objectview=None):
        return (self.pretty_repr(), self.pointer_to_me)
    
    def exportjson(self):
        yield {attr:self[attr] for attr in self.ATTRIBUTES}
        # return {**data, "whittler_filename":self["whittler_filename"]}




from Whittler.classes.RelevanceInterface import RelevanceInterface
from Whittler.classes.RelevanceFilteredResultList import RelevanceFilteredResultList
from Whittler.classes.MemoryCompressor import MaybeCompressedString
from collections import defaultdict, OrderedDict, namedtuple
import re


class SortedResultListDict(defaultdict, RelevanceInterface):

    def __init__(self, default_type, pointer_to_me=None):
        defaultdict.__init__(self, default_type)
        RelevanceInterface.__init__(self, pointer_to_me)
        self._pointer_to_me = pointer_to_me
    
    ChildView = namedtuple("ChildView",[
        "child_key_abridged",
        "child_key_length",
        "total_child_result_count",
        "relevant_child_result_count",
        "child_value",
        "child_pointer"
    ])

    _numeric_tokenize = re.compile(r'(\d+)|(\D+)').findall
    def _numeric_sortkey(self,str):
        return tuple(int(num) if num else alpha for num,alpha in self._numeric_tokenize(str))

    def sorted_childviews(self, objectview=None):
        if objectview is None:
            objectview = self.objectview
        child_repr_values = {}
        for key, value in defaultdict.items(self):
            keylen = str(len(key)) # it needs to be a string to work with _numeric_sortkey
            abridged_key = key.value[:1000] # these can be huge and we won't see all of it anyway
            child_repr_values[key] = self.ChildView(
                child_key_abridged = abridged_key ,
                child_key_length = keylen ,
                total_child_result_count = value.real_length() ,
                relevant_child_result_count = len(value) ,
                child_value = value ,
                child_pointer = self.pointer_to_me.copy().get_by_index(key)
            )
        
        # "reversed" actually yields the expected (i.e. non-reversed-seeming) result...
        reverse = lambda iter: reversed(iter) if objectview["sort_reverse"] else iter
        if objectview["sort_by"] == "total":
            return OrderedDict(reverse(sorted(child_repr_values.items(),
                            key=lambda tup: tup[1].total_child_result_count)))
        elif objectview["sort_by"] == "length":
            if objectview["sort_numeric"]:
                return OrderedDict(reverse(sorted(child_repr_values.items(),
                                key=lambda tup: self._numeric_sortkey(tup[1].child_key_length))))
            else:
                return OrderedDict(reverse(sorted(child_repr_values.items(),
                                key=lambda tup: tup[1].child_key_length)))
        elif objectview["sort_by"] == "attribute value":
            if objectview["sort_numeric"]:
                return OrderedDict(reverse(sorted(child_repr_values.items(),
                                key=lambda tup: self._numeric_sortkey(tup[1].child_key_abridged))))
            else:
                return OrderedDict(reverse(sorted(child_repr_values.items(),
                                key=lambda tup: tup[1].child_key_abridged)))
        else: # elif self._sort_by == "relevant":
            return OrderedDict(reverse(sorted(child_repr_values.items(),
                            key=lambda tup: tup[1].relevant_child_result_count)))
    
    def items(self):
        for k,v in self.sorted_childviews().items():
            yield (k,v.child_value)
        
    def keys(self):
        yield from (k for k,v in self.items())
        
    def values(self):
        yield from (v for k,v in self.items())
    
    def __setitem__(self, key, value):
        assert isinstance(value, RelevanceInterface)
        key = MaybeCompressedString(key)
        child_pointer = self.pointer_to_me.copy().get_by_index(key)
        value.pointer_to_me = child_pointer
        defaultdict.__setitem__(self,key,value)
    
    def __getitem__(self, key):
        return defaultdict.__getitem__(self, MaybeCompressedString(key))
    
    def __delitem__(self, key):
        return defaultdict.__delitem__(self, MaybeCompressedString(key))

    def __len__(self):
        return len([obj for obj in defaultdict.values(self) if obj.relevant])
    

    #######################################
    #  RelevanceInterface implementations
    #
    
    def real_iter_values(self):
        yield from (self[key] for key in defaultdict.__iter__(self))
    

    #################################################
    #  NestedObjectPointerInterface implementations
    #
    
    def give_child_pointers(self):
        return OrderedDict((k,v.child_pointer) for k,v in self.items())
    
    def size(self):
        return sum(obj.size() for obj in self.values())
    
    def all_result_objects(self):
        for resultcontainer in self.values():
            yield from resultcontainer.all_result_objects()
    
    def show_view(self, ct=0, objectview=None):
        if objectview is None:
            objectview = self.objectview
        max_width = self.Config.MAX_OUTPUT_WIDTH
        ret = OrderedDict()
        lines = []
        for key,cv in self.sorted_childviews(objectview=objectview).items():
            value = cv.child_value
            ptr = cv.child_pointer
            if not objectview["show_irrelevant"] and not len(value):
                continue
            # convert each value to bytes and remove the "b''" to escape newlines and control chars etc
            sanitized_key = repr(key.value.strip().encode('utf-8'))[2:-1]
            # Then replace each newline and tab character with three spaces, for visual ease
            sanitized_key = sanitized_key.replace("\\n","   ").replace("\\t","   ")
            line = "| {:4d} | {:5d} | {:8d} | {:8d} | {}".format(ct, cv.total_child_result_count,
                        cv.relevant_child_result_count, int(cv.child_key_length), sanitized_key)
            if len(line) > max_width-2:
                line = line[:max_width-2-7]+" [...] "
            lines.append(line)
            ret[ct] = ptr
            ct += 1
            if not objectview["limit"] is None and len(lines) >= objectview["limit"]:
                break
        headertext = "|  id  | total | relevant |  length  | attribute value  "
        max_linelen = max(len(line) for line in lines) if len(lines) else 0
        max_linelen = len(headertext) if len(headertext) > max_linelen else max_linelen
        header_footer = "+" + "="*(max_linelen-1) + "+\n"
        divider = "|{:-<6}+{:-<7}+{:-<10}+{:-<10}+".format("","","","")
        divider = divider + "-"*(max_linelen-len(divider)) + "|\n"
        header = header_footer
        header += headertext + " "*(max_linelen-len(headertext)) + "|\n"
        header += divider
        s = header
        for line in lines:
            s += line + " "*(max_linelen-len(line)) + "|\n"
        s += header_footer
        return (s, ret)
    
    def exportjson(self):
        for value in self.values():
            if value.relevant:
                yield from value.exportjson()
    

class ResultDict(SortedResultListDict):
    def __init__(self, parent_rdc=None, pointer_to_me=None):
        SortedResultListDict.__init__(self, default_type=RelevanceFilteredResultList, pointer_to_me=pointer_to_me)
        self.parent_rdc = parent_rdc


class ResultDictContainer(SortedResultListDict):
    def __init__(self, resultdb, pointer_to_me=None):
        SortedResultListDict.__init__(self, default_type=lambda: ResultDict(parent_rdc=self), pointer_to_me=pointer_to_me)
        self.resultdb = resultdb



from Whittler.classes.input_utils import *


class ArgumentBase:
    name = NotImplemented
    optional = False

    def __init__(self):
        self.resultdb = None
    
    def __repr__(self):
        if self.optional:
            return f"[[{self.name}]]"
        return f"[{self.name}]"
    
    def __str__(self):
        return f"[{self.name}]"

    # returns None on failure
    def interpret_arg(self, arg, quiet=False):
        raise NotImplementedError()

    def set_name(self, newname):
        self.name = newname
        return self
    
    def set_optional(self):
        self.optional = True
        return self

class ContextPointerArgument(ArgumentBase):
    name = "row"
    def interpret_arg(self, arg, quiet=False):
        if not isinstance(self.resultdb.context_pointers, dict):
            if not quiet:
                wprint(f"Can't dig deeper.")
            return None
        
        try:
            choice = int(arg)
            # The value supplied was an int, so should be looked up in the context_pointers dict
            if choice not in self.resultdb.context_pointers:
                if not quiet:
                    wprint(f"Could not recognize {choice} as one of the IDs listed above.\n")
                return None
            return self.resultdb.context_pointers[choice]
        except ValueError:
            # we couldn't parse it as an int, so it must be a string literal attribute value
            choice = arg
            for ptr in self.resultdb.context_pointers.values():
                # The attribute values will be the .value property of the Vertex object given by the last
                # get_by_index operation called on this pointer.
                if ptr.path[-1].value == arg:
                    return ptr
            if not quiet:
                wprint(f"Could not recognize \"{choice}\" as one of the attributes of this dataset.\n")
            return None
        

class ResultAttributeArgument(ArgumentBase):
    name = "attr"

class IntegerArgument(ArgumentBase):
    name = "int"
    def interpret_arg(self, resultdb, arg):
        try:
            argval = int(arg)
        except ValueError:
            return None
        return argval

class StringArgument(ArgumentBase):
    name = "str"

class FilenameArgument(ArgumentBase):
    name = "fname"


class CommandBase:
    name = NotImplemented
    arguments = []

    def __init__(self, resultdb):
        self.resultdb = resultdb
        for arg in self.arguments:
            arg.resultdb = resultdb

    def execute(self, *args):
        raise NotImplementedError()

    def interpret_args(self, *args):
        ret = []
        for i in range(len(self.arguments)):
            expected_argtype = self.arguments[i]
            if i >= len(args) and not expected_argtype.optional:
                return False
            ret.append(expected_argtype.interpret_arg(args[i]))
        return ret
    



class SingletonMeta(type):
    _singleton_instances = {}
    def __call__(cls, *args, **kwargs):
        if cls not in cls._singleton_instances:
            cls._singleton_instances[cls] = super(SingletonMeta, cls).__call__(*args, **kwargs)
        return cls._singleton_instances[cls]

class Singleton(metaclass=SingletonMeta):
    # Normally this __new__ wouldn't be necessary, but it is required in order to make unpickled objects
    # subscribe to the singleton contract.
    def __new__(cls, *args, **kwargs):
        if cls not in cls._singleton_instances:
            cls._singleton_instances[cls] = super().__new__(cls, *args, **kwargs)
        return cls._singleton_instances[cls]


from Whittler.classes.NestedObjectPointer import NestedObjectPointerInterface


class RelevanceInterface(NestedObjectPointerInterface):

    @property
    def relevant(self):
        return any(result.relevant for result in self.real_iter_values())

    def mark_irrelevant(self):
        for obj in self.real_iter_values():
            obj.mark_irrelevant()

    def mark_relevant(self):
        for obj in self.real_iter_values():
            obj.mark_relevant()

    def real_length(self):
        return len([e for e in self.real_iter_values()])

    def real_iter_values(self):
        raise NotImplementedError()




from Whittler.classes.NestedObjectPointer import NestedObjectPointer
from Whittler.config import Config
import textwrap
import random
from collections import Counter


actions = {
    "general" : {
        "exit" : "Gracefully exit the program",
        "shell" : "Drop into an IPython shell to interact with the dataset."
    },
    "navigation" : {
        "show [[limit]]" : "Show the current data context, up to [limit] entries (shows all entries by default). Mutes "+\
                           "irrelevant results or table entries with 0 relevant results.",
        "showall [[limit]]" : "Show the current data context, up to [limit] entries (shows all entries by default). Includes "+\
                              "irrelevant results or table entries with 0 relevant results.",
        "dig [attr]" : "Dig into a specific data grouping category, either by attribute name, or by attribute id",
        "up" : "Dig up a level into the broader data grouping category",
        "top" : "Dig up to the top level",
        "dump [[limit]]" : "Display every relevant result in every category, up to [limit] entries (shows all by default)",
        "exit" : "Gracefully exit the program"
    },
    "data model interaction" : {
        "irrelevant [[id]]" : "Mark all elements in the current context, or those referenced by [id], as irrelevant results",
        "relevant [[id]]" : "Mark all elements in the current context, or those referenced by [id], as relevant results",
        "filter [str] [attr]" : "Mark all results containing [str] in a particular attribute [attr] as irrelevant "+\
                                "(case-insensitive)",
        "invfilter [str] [attr]" : "Mark all results not containing [str] in a particular attribute [attr] as irrelevant "+\
                                   "(case-insensitive)",
        "find [str] [attr]" : "Find and display all results that contain [str] in their [attr] attribute (case-sensitive)",
        "invfind [str] [attr]" : "Find and display all results that do not contain [str] in their [attr] attribute (case-sensitive)",
        "search [str] [attr]" : "Create a new result group with all results that contain [str] in their [attr] "+\
                                "attribute (case-sensitive)",
        "invsearch [str] [attr]" : "Create a new result group with all results that do not contain [str] in their [attr] "+\
                                   "attribute (case-sensitive)",
        #"group [attr]" : "",
        "fuzzygroup [id] [attr]" : "Using data science, group all results in the database by similarity to the attribute "+\
                                   "[attr] of the result referenced by [id].",
        "ungroup [id]" : "Remove the result referenced by [id] from the result group in the current view.",
        "game [[id]]" : "Play a game where individual results are presented one-by-one, and the user is asked whether "+\
                        "the result is relevant or not and why. Using this information, other similar results are also "+\
                        "eliminated in bulk. If [id] is specified, then the results presented are limited to the result "+\
                        "group represented by the specified [id]."
    },
    "output" : {
        "quiet [attr]" : "Suppress an attribute from being displayed when printing out raw result data",
        "unquiet [attr]" : "Undo the suppression from an earlier quiet command",
        "solo [attr]" : "Print only a single attribute's value when printing out raw result data",
        "SOLO [attr]" : "Print ONLY a single attribute's value when printing out raw result data, with no context IDs or "+\
                        "attribute value names",
        "unsolo" : "Disable solo mode. Note that this retains any attributes suppressed using the \"quiet\" command.",
        "sort [colname]" : "Sorts the displayed results by the value in the specified column. Use quotes if the column name "+\
                           "has a space in it.",
        "sortn [colname]" : "Sorts the displayed results numerically by the value in the specified column. Use quotes if the "+\
                            "column name has a space in it.",
        "rsort [colname]" : "Reverse-sorts the displayed results by the value in the specified column. Use quotes if the column name "+\
                            "has a space in it.",
        "rsortn [colname]" : "Reverse-sorts the displayed results numerically by the value in the specified column. Use quotes if the "+\
                             "column name has a space in it.",
        "history" : "Print all commands that have been run in this session so far",
        "width [numchars]" : "Modify the maxiumum terminal width, in characters, that all output will be formatted to",
        "exportjson [fname] [[id]]" : "Export all relevant results in JSON form into the gzip-compressed file [fname]. "+\
                                      "Optionally, limit the output to the result set as referenced by [id].",
        "export [fname] [[id]]" : "Export all relevant results in Python Pickle (serialized binary) form at into the gzip-compressed "+\
                                  "file [fname]. Optionally, limit the output to the result set as referenced by [id]."
    }
}

# A simple history of the commands that have been run in the Whittler shell.
command_history = []

# The file to which the command history will be written, if any. By default, a .txt file named after
# the current date and time, placed in the .whittler directory in the user's home folder.
command_redirect_file = None
def set_command_redirect_file(obj):
    global command_redirect_file
    command_redirect_file = obj
def get_command_redirect_file():
    return command_redirect_file

# The file to which all Whittler output will be written, if any. By default, a .txt file named after
# the current date and time, placed in the .whittler directory in the user's home folder.
global_redirect_file = None
def set_global_redirect_file(obj):
    global global_redirect_file
    global_redirect_file = obj
def get_global_redirect_file():
    return global_redirect_file

# The file to which the output of the specified command will be written. This adds support for shell-
# style redirection, e.g. "showall > ./my_file.txt"
redirect_file = None
def set_redirect_file(obj):
    global redirect_file
    redirect_file = obj
def get_redirect_file():
    return redirect_file


class _NoInput:
    pass
def wprint(s=_NoInput,end="\n",quiet=False):
    if s is _NoInput:
        s = ""
    if not redirect_file is None:
        redirect_file.write(str(s)+end)
    if not global_redirect_file is None:
        global_redirect_file.write(str(s)+end)
    if not quiet:
        print(s,end=end)

def winput(msg):
    user_input = input(msg)
    command_history.append(user_input)
    if not command_redirect_file is None:
        command_redirect_file.write(user_input+"\n")
    if not redirect_file is None:
        redirect_file.write(str(msg))
        redirect_file.write(user_input+"\n")
    if not global_redirect_file is None:
        global_redirect_file.write(str(msg))
        global_redirect_file.write(user_input+"\n")
    return user_input

def print_help():
    wprint()
    printstr = "|   {{:<{}}}  :  "
    for actioncategory, actiondict in actions.items():
        longest_action_name = max(len(action) for action in actiondict.keys())
        wprint(f"{actioncategory}:")
        for actionname, action_description in actiondict.items():
            prologue = printstr.format(longest_action_name).format(actionname)
            description = textwrap.wrap(action_description, Config.MAX_OUTPUT_WIDTH-len(prologue))
            for line in description:
                wprint(f"{prologue}{line}")
                prologue = "|" + " "*(len(prologue)-1)
        wprint()
    wprint("NOTE: This shell supports quoted arguments and redirecting command output to a file using the \">\" operator.")
    wprint()

# False means no value was supplied for this arg position, None means failed to parse as int
def get_int_from_args(args, position=0):
    if not len(args) > position:
        return False
    else:
        try:
            argval = int(args[position])
        except ValueError:
            return None
    return argval

# False means no value was supplied for this arg position, None means failed to find corresponding context pointer
def get_ptr_from_id_arg(resultdb, args, id_arg_position=0, quiet=False):
    if not isinstance(resultdb.context_pointers, dict):
        if not quiet:
            wprint(f"Can't dig deeper.")
        return None
    
    choice = get_int_from_args(args, position=id_arg_position)
    
    # No value was supplied for this arg position
    if choice is False:
        return False
    
    # we couldn't parse it as an int, so it must be a string literal attribute value
    if choice is None:
        choice = args[id_arg_position]
        for ptr in resultdb.context_pointers.values():
            # The attribute values will be the .value property of the Vertex object given by the last
            # get_by_index operation called on this pointer.
            if ptr.path[-1].value == choice:
                return ptr.copy()
        if not quiet:
            wprint(f"Could not recognize \"{choice}\" as one of the attributes of this dataset.\n")
        return None
    
    # The value supplied was an int, so should be looked up in the context_pointers dict
    elif choice not in resultdb.context_pointers:
        if not quiet:
            wprint(f"Could not recognize {choice} as one of the IDs listed above.\n")
        return None
    return resultdb.context_pointers[choice]

# False means no value was supplied for this arg position, None means failed to find corresponding context pointer
def get_attrname_from_attribute_arg(resultdb, args, attr_arg_position=0):
    # ptr = get_ptr_from_id_arg(resultdb, args, id_arg_position=attr_arg_position, quiet=True)
    # if ptr is False:
    #     return False
    attrname = args[attr_arg_position]
    if attrname not in resultdb.result_class.ATTRIBUTES:
        return None
    return attrname

def select_attribute(resultdb, msg):
    attrs = resultdb.result_class.ATTRIBUTES
    for i in range(len(attrs)):
        wprint(f" {i} : {attrs[i]}")
    wprint()
    try:
        index = int(winput(msg))
    except ValueError:
        return None
    return attrs[index]

def group_interactive(resultdb, groupattr, groupval, max_print_count=10, max_print_chars_per_attrval=1000):
    wprint("Creating a result group set based on this attribute value...")
    SIMILARITY_THRESHOLD = resultdb.Config.SIMILARITY_THRESHOLD
    wprint("Using similarity threshold {:.2f}...".format(SIMILARITY_THRESHOLD))
    result_similarities = resultdb.find_similar_results(groupattr, groupval)
    while True:
        similar_results = list(res for res,similarity in result_similarities if similarity>SIMILARITY_THRESHOLD)
        attrvals_of_similar_results = Counter()
        for res in similar_results:
            attrval = dict.__getitem__(res,groupattr)
            attrvals_of_similar_results[attrval] += 1
        ordered_attrvals_of_similar_results = attrvals_of_similar_results.most_common()[:max_print_count]
        for attrval_mcs,count in ordered_attrvals_of_similar_results:
            attr_printstr = attrval_mcs.value
            if len(attr_printstr) > max_print_chars_per_attrval:
                attr_printstr = attr_printstr[:max_print_chars_per_attrval] + " [...]"
            wprint(f"----- {count} result{'s' if count>1 else ''} with value:\n{attr_printstr}\n")
        if max_print_count < len(attrvals_of_similar_results):
            wprint(f"\n... and {sum(attrvals_of_similar_results.values())-sum(count for attrval,count in ordered_attrvals_of_similar_results)} more ...\n")
        wprint(f"I found {len(similar_results)} similar results with \"{groupattr}\" values as shown above.")
        wprint("I want to create a result group set based on these findings. Is this grouping...")
        wprint("   OK? (1)")
        wprint("   too conservative? (2)")
        wprint("   too liberal? (3)")
        wprint("   never mind, abandon group creation (4)")
        wprint("   not sure, show more results (5)")
        while True:
            grouping_choice = winput("> ")
            bad_input = False
            try:
                grouping_choice = int(grouping_choice)
                if grouping_choice not in [1,2,3,4,5]:
                    bad_input = True
            except:
                bad_input = True
            if bad_input:
                wprint("I didn't understand your input.")
                continue
            break
        if grouping_choice == 1:
            pass
        elif grouping_choice == 2:
            SIMILARITY_THRESHOLD -= SIMILARITY_THRESHOLD / Config.SIMILARITY_THRESHOLD_MODIFICATION_FACTOR
            wprint("Using similarity threshold {:.2f}...".format(SIMILARITY_THRESHOLD))
            continue
        elif grouping_choice == 3:
            SIMILARITY_THRESHOLD += (1 - SIMILARITY_THRESHOLD) / Config.SIMILARITY_THRESHOLD_MODIFICATION_FACTOR
            wprint("Using similarity threshold {:.2f}...".format(SIMILARITY_THRESHOLD))
            continue
        elif grouping_choice == 4:
            wprint("OK, abandoning group creation.")
            break
        elif grouping_choice == 5:
            max_print_count *= 2
            max_print_chars_per_attrval *= 2
            continue
        resultdb.register_grouped_results(groupattr, groupval, similar_results)
        wprint("Created result group based on this entry. You can view this group in the \"groups\" pane in the top context.")
        wprint("(You might have to use the \"showall\" command, since all results in this group are now marked irrelevant.)")
        ptr = NestedObjectPointer(resultdb)
        ptr.access_property("grouped_results")
        ptr.get_by_index(groupval)
        return ptr

def parse_user_input(user_input):
    if not user_input.strip():
        return None
    
    # Support strings with escaped quote characters
    user_input = user_input.replace("\\'","###_TRUFFLEHOG_SINGLE_QUOTE_###")
    user_input = user_input.replace("\\\"","###_TRUFFLEHOG_DOUBLE_QUOTE_###")

    # Support multiple commands in a single line with semicolons, but also ignore escaped semicolons
    user_input = user_input.replace("\\;","__WHITTLER_SEMICOLON_ESCAPE_W89UHA938F__")

    inputs = [cmd.strip().replace("__WHITTLER_SEMICOLON_ESCAPE_W89UHA938F__",";") for cmd in user_input.split(";")]

    ret = []
    for inp in inputs:
        # For now, just treat single and double quotes identically.
        inp = inp.replace("'","\"")

        if not inp.strip():
            continue

        if inp.find("\"") == -1:
            verb,*args = list(filter(None,inp.split(" ")))
        else:
            if inp.count("\"") % 2:
                wprint("Failed to parse quoted input.")
                return None
            quote_regions = inp.split("\"")
            # Obviously we're not going to support nested quotes, so every other quote_region will be
            # a string contained by quotes. Ideally I should be using the shlex library for this but
            # I'm too lazy.
            all_args = []
            for i in range(len(quote_regions)):
                if not i%2:
                    all_args.extend(list(filter(None,quote_regions[i].split(" "))))
                else:
                    all_args.append(quote_regions[i])
            verb,*args = all_args
        try:
            redirect_to_file_index = args.index(">")
            if redirect_to_file_index != len(args)-2:
                wprint("Specify only a single file to which to redirect command output! (Maybe try absolute paths; quotes are supported.)")
                return None
            redirect = args[-1]
            args = args[:-2]
        except ValueError:
            redirect = None
        
        for i in range(len(args)):
            args[i] = args[i].replace("###_TRUFFLEHOG_SINGLE_QUOTE_###","\\'")
            args[i] = args[i].replace("###_TRUFFLEHOG_DOUBLE_QUOTE_###","\\\"")
        ret.append((verb,args,redirect))
    return ret

irrelevance_filters = []
already_marked_relevant = []
def play_elimination_game(resultdb, obj):
    global irrelevance_filters
    global already_marked_relevant
    game_actions = {
        1 : "mark as relevant",
        2 : "mark as irrelevant",
        3 : "mark as ambiguous",
        4 : "quit game",
        5 : "clear relevancy filters"
    }
    def filterfunc(result):
        for attr, value in irrelevance_filters:
            if result[attr].strip() == value.strip():
                return False
        return True
    while True:
        results = list(filter(filterfunc, obj.all_result_objects()))
        if not len(results):
            wprint("All results accounted for - Game Over :)")
            break
        random_result = random.choice(results)
        wprint("\n".join(["| "+line for line in random_result.show_view()[0].splitlines()]))
        question = ""
        for action_id, action_description in game_actions.items():
            question += f" {action_id} : {action_description}\n"
        question += "\naction? "
        answer = winput(question)
        try:
            answer = int(answer)
            if not answer in game_actions.keys():
                wprint("Unrecognized selection.")
                continue
        except ValueError:
            wprint("Unrecognized selection.")
            continue
        if answer == 4:
            wprint("\nEnding game, thanks for playing.")
            break
        elif answer == 3:
            wprint("\nOK, taking no action.")
            continue
        elif answer == 1:
            wprint("\nOK, taking no action.")
            continue # todo
            relevant_attr = select_attribute(resultdb, "Which specific value makes this result definitely relevant? ")
            irrelevance_filters.append((relevant_attr,random_result[relevant_attr]))
            wprint("OK, I'll ignore results with that value for that attribute for the rest of the game.")
            continue
        elif answer == 5:
            wprint("Currently, you have relevancy filters on the following attributes:")
            for attr, value in irrelevance_filters:
                wprint(attr)
            confirm = winput("Are you sure you want to clear these relevancy filters? (Y/n) ")
            if confirm.lower() == "n":
                irrelevance_filters = []
                wprint("Cleared.")
                continue
        # answer must be 2 (irrelevant) past this point.
        wprint("\nOK, this result will be marked as irrelevant.\n")
        random_result.mark_irrelevant()
        num_results_eliminated = 1
        wprint("I can try to find other results similar to this one through various heuristics, and mark them irrelevant as well.\n")

        search_actions = {
            1 : "contains the specific value above",
            2 : "contains a particular substring (case-insensitive)",
            3 : "loosely resembles this value (fuzzy string matching)",
            4 : "never mind"
        }
        initial_str = "OK. To start, which"
        again_str = "Search for similar results? (y/N) "
        while True:
            go_again = winput(again_str)
            if go_again is None:
                wprint("Unrecognized input.")
                continue
            again_str = "Search again? (y/N)"
            if not go_again.strip() or go_again.lower() != "y":
                break
            wprint()
            wprint("\n".join(["| "+line for line in random_result.show_view()[0].splitlines()]))
            wprint()
            problematic_attr = select_attribute(resultdb, f"{initial_str} attribute's value makes this result irrelevant? ")
            initial_str = "Which"
            if not problematic_attr:
                wprint("Unrecognized selection.")
                continue
            wprint(f"\n\nFor reference, here is the value of the \"{problematic_attr}\" attribute for this result:")
            wprint("_"*Config.MAX_OUTPUT_WIDTH)
            wprint(random_result[problematic_attr])
            wprint("^"*Config.MAX_OUTPUT_WIDTH)

            question = f"\nI can search for results where the \"{problematic_attr}\" attribute:\n"
            for action_id, action_description in search_actions.items():
                question += f" {action_id} : {action_description.format(problematic_attr)}\n"
            question += "\naction? "
            answer = winput(question)
            try:
                answer = int(answer)
                if not answer in search_actions.keys():
                    wprint("Unrecognized selection.")
                    continue
            except ValueError:
                wprint("Unrecognized selection.")
                continue
            generate_report = True
            if answer == 1:
                irrelevant_resultlist = resultdb.categorized_results[problematic_attr][random_result[problematic_attr]]
                num_results_eliminated = len(irrelevant_resultlist)
                irrelevant_resultlist.mark_irrelevant()
            elif answer == 2:
                substring = winput("What is the substring you would like to filter by? ")
                if substring.lower() not in random_result[problematic_attr].lower():
                    wprint(f"Sorry, that is not a case-insensitive substring of this result's \"{problematic_attr}\" attribute...")
                    continue
                irrelevant_resultlist = resultdb.categorized_results[problematic_attr][random_result[problematic_attr]]
                for result in resultdb.results:
                    if substring.lower() in result[problematic_attr].lower():
                        result.mark_irrelevant()
                        num_results_eliminated += 1
            elif answer == 3:
                group_ptr = group_interactive(resultdb, problematic_attr, random_result[problematic_attr])
                if group_ptr is None:
                    continue
                irrelevant_resultdict = group_ptr.give_pointed_object()
                num_results_eliminated = len([r for r in irrelevant_resultdict.all_result_objects()])
                irrelevant_resultdict.mark_irrelevant()
            elif answer == 4:
                generate_report = False
            
            if generate_report:
                excitement = num_results_eliminated//100
                report = f"\n > Eliminating {num_results_eliminated} results{'.' if not excitement else '!'*excitement}\n"
                if excitement >= 3:
                    report = report.upper()
                wprint(report)



from Whittler.classes.RelevanceInterface import RelevanceInterface
from collections import OrderedDict
import re


class RelevanceFilteredResultList(list, RelevanceInterface):
    def __init__(self, iterable=None, pointer_to_me=None):
        list.__init__(self)
        RelevanceInterface.__init__(self, pointer_to_me)
        if not iterable is None:
            for e in iterable:
                self.append(e)
    
    def __setitem__(self, index, value):
        assert isinstance(value, RelevanceInterface)
        list.__setitem__(self,index,value)
    
    def __iter__(self):
        yield from (e for e in self.real_iter_values() if e.relevant)
    
    def __len__(self):
        return len([e for e in self])
    
    # todo: speed up "result in RFRLobj" lookups - most of the time is spent checking "if result in self.results" in ResultDatabase.add_result
    
    def append(self, newitem):
        assert isinstance(newitem, RelevanceInterface)
        list.append(self,newitem)
    
    def extend(self, iterable):
        for e in iterable:
            self.append(e)
    
    def yield_irrelevant(self):
        yield from (e for e in list.__iter__(self) if not e.relevant)
    
    def get_by_id(self, result_id):
        return list(filter(self.real_iter_values(),key=lambda result:str(hash(result))[:6] == result_id))[0]
    

    #######################################
    #  RelevanceInterface implementations
    #
    
    def real_iter_values(self):
        yield from list.__iter__(self)


    #################################################
    #  NestedObjectPointerInterface implementations
    #
    
    def give_child_pointers(self):
        ret = OrderedDict()
        for i in range(self.real_length()):
            if self[i].relevant:
                ret[i] = self.pointer_to_me.copy().get_by_index(i)
        return ret
    
    def size(self):
        return len(self)
    
    def all_result_objects(self):
        yield from (result for result in self)

    _numeric_tokenize = re.compile(r'(\d+)|(\D+)').findall
    def _numeric_sortkey(self,str):
        return tuple(int(num) if num else alpha for num,alpha in self._numeric_tokenize(str))

    def show_view(self, ct=0, objectview=None):
        if objectview is None:
            objectview = self.objectview
        ret = OrderedDict()
        s = ""
        def adds(string, prefix=""):
            nonlocal s
            s += prefix + string + "\n"
        child_ptrdict = self.give_child_pointers()
        if not objectview["sort_by"] is None:
            # "reversed" actually yields the expected (i.e. non-reversed-seeming) result...
            reverse = lambda iter: reversed(iter) if objectview["sort_reverse"] else iter
            if objectview["sort_numeric"]:
                child_ptrdict = OrderedDict(reverse(sorted(child_ptrdict.items(),
                    key=lambda tup: self._numeric_sortkey(self[tup[0]][objectview["sort_by"]]))))
            else:
                child_ptrdict = OrderedDict(reverse(sorted(child_ptrdict.items(),
                    key=lambda tup: self[tup[0]][objectview["sort_by"]])))
        for key, ptr in child_ptrdict.items():
            result = self[key]
            if not objectview["show_irrelevant"] and not result.relevant:
                continue
            result_id = str(hash(result))[:9]
            if not objectview["SOLO"] is None:
                adds(result.pretty_repr())
            else:
                adds("~~~~~~~~~~~~~~~~~~~~~~~~~")
                adds(f"  Result with ID {result_id}    ")
                adds("~~~~~~~~~~~~~~~~~~~~~~~~~")
                for line in result.pretty_repr().split("\n"):
                    adds(line, prefix="| ")
                s += "\n\n"
            ret[int(result_id)] = ptr
            if not objectview["limit"] is None and len(ret) >= objectview["limit"]:
                break
        return (s, ret)
    
    def exportjson(self):
        for result in self:
            yield from result.exportjson()
    



from Whittler.classes.Singleton import Singleton

WHITTLER_VERSION = "1.1.3"

class Metadata(Singleton):
    
    version = WHITTLER_VERSION

    def major_versions_match(self):
        return self.version.split(".")[0] == WHITTLER_VERSION.split(".")[0]

    def minor_versions_match(self):
        return ".".join(self.version.split(".")[:2]) == ".".join(WHITTLER_VERSION.split(".")[:2])

    def revision_versions_match(self):
        return self.version == WHITTLER_VERSION

    def to_dict(self):
        # get all instance variables that aren't dunder variables like __class__
        ret = {attrname:getattr(self,attrname) for attrname in filter(lambda name: not name.startswith("__"), dir(self))}
        
        # remove all callable functions specified in this exact class definition (like exportjson())
        return dict(filter(lambda tup: not callable(getattr(type(self),tup[0],None)), ret.items()))

    def __repr__(self):
        md = self.exportjson()
        longest_metadata_varname_length = max(len(varname) for varname in md.keys())
        fmt = f"   {{:>{longest_metadata_varname_length}}} : {{}}\n"
        s = "WHITTLER METADATA:\n"
        for k,v in self.to_dict().items():
            # convert each value to bytes and remove the "b''" to escape newlines and control chars etc
            s += fmt.format(k, repr(v.encode('utf-8'))[2:-1])
        return s

    def exportjson(self):
        # convert non-string values to their string representations for storage in JSON
        return {key:f"{value}" for key,value in self.to_dict().items()}

metadata_only_instance = Metadata()

class MetadataInterface:
    metadata = metadata_only_instance


# from Whittler.config import ConfigurableInterface
from Whittler.classes.Singleton import Singleton
import zlib

# TODO: this can all be coded in cpython! That would be a kickass library to have exist!
# https://docs.python.org/3/c-api/buffer.html#bufferobjects

# based on https://stackoverflow.com/questions/479218/how-to-compress-small-strings
class MemoryCompressor(Singleton):#, ConfigurableInterface):
    def __init__(self):
        # wbits=-15 gives a raw output stream with no header or checksum, with a 2**15 byte window.
        self.compressor = zlib.compressobj(wbits=-15)
        self.decompressor = zlib.decompressobj(wbits=-15)
        self.junk_offset = 0

        self.trainee_compression_callbacks = []
        self.training_mode = True
        self.total_train_count = 100000

        self.COMPRESSION_ENABLED = False

    def train(self, training_str):
        assert self.training_mode
        training_bytes = training_str.encode('utf-8')
        self.junk_offset += len(training_bytes)

        # run the training line through the compressor and decompressor
        self.junk_offset -= len(self.decompressor.decompress(self.compressor.compress(training_bytes)))

        # use Z_SYNC_FLUSH. A full flush seems to detrain the compressor, and 
        # not flushing wastes space.
        self.junk_offset -= len(self.decompressor.decompress(self.compressor.flush(zlib.Z_SYNC_FLUSH)))

        # If we have trained on at least self.total_train_count result objects, then disable training_mode.
        # The second half of the if statement is necessary because we call train() on each call to Result.__setitem__,
        # not on each creation of a Result object.
        if len(self.trainee_compression_callbacks) >= self.total_train_count:
            print("\nMEMORY COMPRESSION TRAINING DONE.")
            self.disable_training_mode()
            return self.compress(training_str.encode('utf-8'))
        
        return training_str
    
    def add_compression_callback(self, mcd):
        self.trainee_compression_callbacks.append(mcd)
    
    def disable_training_mode(self):
        self.training_mode = False
        for mcd in self.trainee_compression_callbacks:
            assert not mcd.compressed
            compressed = self.compress(mcd)
            del mcd[:]
            mcd.extend(compressed)
            mcd.compressed = True

    # Takes bytearray or bytes, returns bytes
    def compress(self,b):
        compressor = self.compressor.copy()
        ret = compressor.compress(b)+compressor.flush()
        return ret

    # Takes bytearray or bytes, returns bytes
    def decompress(self,b):
        decompressor = self.decompressor.copy()
        return (decompressor.decompress(b)+decompressor.flush())[self.junk_offset:]

MemoryCompressorOnlyInstance = MemoryCompressor()

_MCS_cache = {}

class MaybeCompressedString(bytearray):
    
    def __new__(cls, data_string):
        if type(data_string) == MaybeCompressedString:
            return data_string
        assert type(data_string) == str
        data_string_hash = hash(data_string)
        if data_string_hash in _MCS_cache:
            return _MCS_cache[data_string_hash]
        data_bytes = data_string.encode('utf-8')
        add_callback = False
        if MemoryCompressorOnlyInstance.COMPRESSION_ENABLED:
            if MemoryCompressorOnlyInstance.training_mode:
                add_callback = True
                data = MemoryCompressorOnlyInstance.train(data_string)
                if type(data) is bytes:
                    compressed = True
                else:
                    data = data_bytes
                    compressed = False
            else:
                data = MemoryCompressorOnlyInstance.compress(data_bytes)
                compressed = True
        else:
            data = data_bytes
            compressed = False
        ret = bytearray.__new__(cls)
        ret._cached_hash = data_string_hash
        ret.extend(data)
        ret.compressed = compressed
        if add_callback:
            MemoryCompressorOnlyInstance.add_compression_callback(ret)
        _MCS_cache[data_string_hash] = ret
        return ret
    
    def __init__(self, d):
        pass

    def __reduce__(self):
        cls = self.__class__
        args = (self.value,)
        return (cls,args)
    
    @property
    def value(self):
        if self.compressed:
            return MemoryCompressorOnlyInstance.decompress(self).decode('utf-8')
        return self.decode('utf-8')
    
    def __repr__(self):
        value = self.value
        valuerepr = f"{value[:97]}..." if len(self.value) > 100 else value
        return f"{self.__class__.__name__}({valuerepr})"
    
    __str__ = __repr__

    def __hash__(self):
        cached_hash = self._cached_hash
        if cached_hash:
            return cached_hash
        return hash(self.value)
    
    def __eq__(self, other):
        return hash(self) == hash(other)
    
    def __len__(self):
        return len(self.value)
    



from Whittler.classes.Result import Result
from Whittler.classes.RelevanceInterface import RelevanceInterface
from Whittler.classes.ResultDictContainer import ResultDictContainer
from Whittler.classes.RelevanceFilteredResultList import RelevanceFilteredResultList
from Whittler.classes.input_utils import wprint
from Whittler.config import Config
from collections import OrderedDict
import time
import os
import json
import numpy as np
import pickle
import gzip

try:
    from pyxdameraulevenshtein import normalized_damerau_levenshtein_distance
except ImportError:
    print("Warning: pyxDamerauLevenshtein module not detected, fuzzy grouping logic will be impaired.")
    print("         To install it, run 'pip install pyxDamerauLevenshtein'.")
    normalized_damerau_levenshtein_distance = lambda reference, value: 1.
import difflib

# import cProfile, pstats, io
# from pstats import SortKey


class ResultDatabase(RelevanceInterface):
    
    def __init__(self, result_class=Result, pointer_to_me=None):
        assert issubclass(result_class, Result)
        RelevanceInterface.__init__(self, pointer_to_me)

        self.result_class = result_class

        result_pointer = self.pointer_to_me.copy().access_property("results")
        self.results = RelevanceFilteredResultList(pointer_to_me=result_pointer)

        categorized_results_pointer = self.pointer_to_me.copy().access_property("categorized_results")
        self.categorized_results = ResultDictContainer(self, pointer_to_me=categorized_results_pointer)

        grouped_results_pointer = self.pointer_to_me.copy().access_property("grouped_results")
        self.grouped_results = ResultDictContainer(self, pointer_to_me=grouped_results_pointer)
        
        self.current_pointer = self.pointer_to_me.copy()
        self.context_pointers = OrderedDict()

        # self.pr = cProfile.Profile()
    
    def __getitem__(self, nestedobjectpointer):
        assert nestedobjectpointer.base_object is self
        return nestedobjectpointer.give_pointed_object()

    def add_result(self, result, lookup_set=None):
        assert isinstance(result,self.result_class)
        if lookup_set is None:
            pass
        else:
            result_hash = hash(result)
            if result_hash in lookup_set:
                return
            lookup_set.add(result_hash)
        result._frozen = True
        self.results.append(result)
        for attr in self.result_class.ATTRIBUTES:
            attrval = dict.__getitem__(result, attr) # gets the underlying CompressedBytes object if memory compression enabled
            self.categorized_results[attr][attrval].append(result)


    #########################
    #  Navigation functions
    #

    def construct_view(self, pointer=None, override_options={}):
        if pointer is None:
            pointer = self.current_pointer
        obj = pointer.give_pointed_object()
        if override_options:
            options = {k:v for k,v in obj.objectview.items()}
            options.update(override_options)
        else:
            options = None
        return obj.show_view(objectview=options)
    
    def update_current_view(self, override_options={}):
        viewstr, ptrcontext = self.construct_view(self.current_pointer, override_options=override_options)
        self.context_pointers = ptrcontext
        return viewstr
    
    def navigate_view(self, pointer=None, override_options={}):
        if pointer is None:
            self.current_pointer = self.pointer_to_me.copy()
        else:
            self.current_pointer = pointer.copy()
        return self.update_current_view(override_options=override_options)


    #######################
    #  Grouping functions
    #
    
    @staticmethod
    def compute_distances(reference,values,exp=None):
        if exp is None:
            exp = Config.SIMILARITY_EXPONENT
        
        sm = difflib.SequenceMatcher()
        sm.set_seq2(reference)
        sm_distances = []
        dl_distances = []
        for val in values:
            sm.set_seq1(val)
            sm_distances.append(sm.ratio())
            dl_distances.append(1-normalized_damerau_levenshtein_distance(reference, val))
        sm_distances = np.array(sm_distances)
        dl_distances = np.array(dl_distances)
        dl_exp = np.power(dl_distances, exp)
        sm_exp = np.power(sm_distances, exp)
        dist_sum = dl_exp+sm_exp
        return np.power(dist_sum,1/exp)
    
    def find_similar_results(self, attrname, groupval):
        if not groupval.strip():
            all_results = (res[attrname].strip() for res in list.__iter__(self.results) if hasattr(res,attrname) and res[attrname].strip())
        else:
            all_results = (res[attrname].strip() for res in list.__iter__(self.results))
        return list(zip(self.results,self.compute_distances(groupval, all_results)))
    
    def register_grouped_results(self, attrname, groupval, result_group):
        for result in result_group:
            attrval = dict.__getitem__(result, attrname) # gets the underlying CompressedBytes object if memory compression enabled
            self.grouped_results[groupval][attrval].append(result)


    #################################
    #  Input file parsing functions
    #

    def parse_from_file(self,fname,hash_cache=None):
        result_dict_list = self.result_class.give_result_dict_list(fname)
        last_report = time.time()
        start = last_report
        ct = 0
        parsing_str = f"PARSING {fname} : "
        longest_str_length = 0
        for resultdict in result_dict_list:
            cur_time = time.time()
            if cur_time-last_report > 5:
                print_str = f"{parsing_str}{(int((ct/len(result_dict_list)*100)))}% done ({ct} out of {len(result_dict_list)})"
                if len(print_str)>longest_str_length:
                    longest_str_length = len(print_str)
                wprint(print_str, end='\r')
                last_report = cur_time
            resultdict["whittler_filename"] = fname
            self.add_result(self.result_class(resultdict), lookup_set=hash_cache)
            ct += 1
        if longest_str_length:
            timing = "{:.2f}".format(time.time()-start)
            final_report = f"{parsing_str}Done (took {timing}s)"
            wprint(final_report + " "*(len(final_report)-max(Config.MAX_OUTPUT_WIDTH,longest_str_length)))

    def parse_from_directory(self,dirname,hash_cache=None,multiprocessing_import=False):
        files = os.listdir(dirname)
        last_report = time.time()
        start = last_report
        ct = 0
        parsing_str = f"FILES PARSED: "
        longest_str_length = 0
        wprint(f"Parsing from {dirname} ...")
        try:
            if multiprocessing_import:
                import multiprocessing as mp
                p = mp.Pool()
                # # do this to support KeyboardInterrupt
                # # https://stackoverflow.com/questions/1408356/keyboard-interrupts-with-pythons-multiprocessing-pool
                # resultdictlist_generator = (resultdict for resultdict in p.map_async(
                #     self.result_class._give_result_dict_list,
                #     (dirname+"/"+fname for fname in files),
                #     chunksize=20).get(999999))
                resultdictlist_generator = p.imap_unordered(
                    self.result_class._give_result_dict_list,
                    (dirname+"/"+fname for fname in files),
                    chunksize=20)
                p.close()
            else:
                resultdictlist_generator = (self.parse_from_file(dirname+"/"+fname, hash_cache=hash_cache) for fname in files)
            for resultdictlist in resultdictlist_generator:
                cur_time = time.time()
                if cur_time-last_report > 5:
                    print_str = f"{parsing_str}{(int((ct/len(files)*100)))}% ({ct} out of {len(files)})"
                    if len(print_str)>longest_str_length:
                        longest_str_length = len(print_str)
                    wprint(print_str, end='\r')
                    last_report = cur_time
                if multiprocessing_import:
                    for result in resultdictlist:
                        self.add_result(result, lookup_set=hash_cache)
                ct += 1
        finally:
            if multiprocessing_import:
                p.terminate()
        timing = "{:.2f}".format(time.time()-start)
        final_report = f"Parsing from {dirname} done (took {timing}s)"
        wprint(final_report + " "*(len(final_report)-max(Config.MAX_OUTPUT_WIDTH,longest_str_length)))
    
    def parse_from_export(self,fname,hash_cache=None):
        importing_str = f"IMPORTING {os.path.basename(fname)} ... "
        last_report = time.time()
        start = last_report
        wprint(f"{importing_str}", end='\r')
        pickle_import = False
        # self.pr.enable()
        try:
            try:
                with gzip.GzipFile(fname, "rb") as f:
                    pickle.load(f)
                def result_gen():
                    with gzip.GzipFile(fname, "rb") as f:
                        while True:
                            try:
                                yield pickle.load(f)
                            except EOFError:
                                break
                results = result_gen()
                results_len = None
                pickle_import = True
            except pickle.UnpicklingError:
                try:
                    with gzip.GzipFile(fname, "rb") as f:
                        results = json.loads(f.read().decode('utf-8'))
                        results_len = len(results)
                except json.decoder.JSONDecodeError:
                    raise Exception("Failed to import file as either binary (pickle) or JSON data.")
        except:
            raise
        # self.pr.disable()
        # If we're importing from JSON, we need to make sure that the data keys are compatible with the specified module
        if not pickle_import:
            first_result_keys = set(results[0].keys())
            for result in results:
                if first_result_keys ^ set(result.keys()): # check symmetric difference, basically ensure that they're equal
                    raise Exception(f"invalid Whittler export file {fname} ... all results dicts in the JSON must have the same "+\
                                    f"set of keys, but the keys in the result:\n{result}\ndid not match the expected keys:\n"+\
                                    f"{first_result_keys}")
            if not isinstance(self.result_class.ATTRIBUTES, list):
                self.result_class.ATTRIBUTES = []
            attrs = self.result_class.ATTRIBUTES
            for key in results[0].keys(): # don't use first_result_keys because the json is ordered whereas sets are not
                if key not in attrs:
                    attrs.append(key)
        ct = 0
        biggest_status_str_len = 0
        for result in results:
            cur_time = time.time()
            if cur_time-last_report > 5:
                if results_len:
                    progress_str = f"{(int((ct/results_len*100)))}% done ({ct} out of {results_len})"
                else:
                    progress_str = f"{ct} done"
                status_str = f"{importing_str}{progress_str}"
                if len(status_str) > biggest_status_str_len:
                    biggest_status_str_len = len(status_str)
                wprint(status_str, end='\r')
                last_report = cur_time
            if pickle_import:
                self.add_result(result, lookup_set=hash_cache)
            else:
                self.add_result(self.result_class(result), lookup_set=hash_cache)
            ct += 1
        # s = io.StringIO()
        # sortby = SortKey.CUMULATIVE
        # ps = pstats.Stats(self.pr, stream=s).sort_stats(sortby)
        # ps.print_stats()
        # print(s.getvalue())
        tot_time = "{:.2f}".format(time.time()-start)
        done_str = f"{importing_str}Done (took {tot_time}s)"
        wprint(done_str+" "*max(biggest_status_str_len,Config.MAX_OUTPUT_WIDTH-len(done_str)))


    #######################################
    #  RelevanceInterface implementations
    #
    
    def real_iter_values(self):
        yield from self.results.real_iter_values()
    

    #################################################
    #  NestedObjectPointerInterface implementations
    #
    
    def give_child_pointers(self):
        ret = OrderedDict()
        
        ptr = self.pointer_to_me.copy().access_property("categorized_results")
        ret["categorized_results"] = ptr
        
        ptr = self.pointer_to_me.copy().access_property("grouped_results")
        ret["grouped_results"] = ptr
        
        return ret
    
    def size(self):
        return self.results.size()
    
    def all_result_objects(self):
        yield from self.results.all_result_objects()
    
    def show_view(self, ct=0, objectview=None):
        if objectview is None:
            objectview = self.objectview
        s = "\n"
        s += "+=================================+\n"
        s += "|  {:6d} total relevant results  |\n".format(len(self.results))
        s += "+=================================+\n"
        ret = OrderedDict()
        for childname, childptr in self.give_child_pointers().items():
            s += f"\n{childname}:\n"
            childobj = childptr.give_pointed_object()
            s2, resultdict = childobj.show_view(ct=ct, objectview=objectview)
            s += s2
            for k,v in resultdict.items():
                ret[k] = v
            ct += len(resultdict)
        return (s, ret)

    def exportjson(self):
        yield from self.results.exportjson()
    































#!/usr/bin/env python

import os
import sys

WHITTLER_DIRNAME = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.dirname(WHITTLER_DIRNAME))
try:
    sys.path.remove(WHITTLER_DIRNAME)
except ValueError: # Already removed
    pass

from Whittler.classes.ResultDatabase import ResultDatabase
from Whittler.classes.Result import Result
from Whittler.classes.input_utils import *
from Whittler.classes.NestedObjectPointer import ObjectView
from Whittler.classes.RelevanceFilteredResultList import RelevanceFilteredResultList
import importlib
import importlib.machinery
from pathlib import Path
import datetime
import argparse
import json
import pickle
import gzip
import traceback

try:
    import IPython
    IPYTHON_INSTALLED = True
except ImportError:
    IPYTHON_INSTALLED = False


def main_loop(resultdb):
    global redirect_file, global_redirect_file, cached_commands
    current_view = None
    while True:
        try:
            if not redirect_file is None:
                redirect_file.close()
                redirect_file = None
            if current_view is None:
                current_view = resultdb.navigate_view()
            if len(cached_commands):
                user_input = cached_commands[0]
                cached_commands = cached_commands[1:]
                wprint(f"Whittler > {' '.join([user_input[0]]+user_input[1])}")
            else:
                user_input = parse_user_input(winput("Whittler > "))
                if user_input is None:
                    continue
                if len(user_input) > 1:
                    cached_commands = user_input[1:]
                user_input = user_input[0]
            if user_input is None:
                continue
            verb,args,redirect = user_input
            if verb.startswith("#"):
                continue

            if not redirect is None:
                try:
                    redirect_file = open(redirect, "w+", encoding="utf-8")
                except PermissionError:
                    wprint("Failed to open the specified file, maybe try an absolute path? (FYI, quotes are supported.)")


            # In the following switch statement, "continue" to skip updating the current view,
            # otherwise, the view will be updated. Set the below navigate_ptr variable to the
            # pointer to which we want to navigate, and it will be reset on the next loop.
            navigate_ptr = None

            if verb == "help":
                print_help()
                continue
            elif verb == "shell":
                if IPYTHON_INSTALLED:
                    wprint("Welcome to the subshell, you can use the resultdb object to interact with the dataset.")
                    # with IPython.utils.io.capture_output() as output:
                    IPython.embed(colors="neutral")
                    # wprint(output.stdout,quiet=True)
                else:
                    wprint("No IPython package detected, falling back to basic eval() REPL.")
                    wprint("Type \"exit\" to exit back to the main Whittler shell.")
                    while True:
                        try:
                            cmd = winput("> ").strip()
                            if cmd == "exit":
                                break
                            wprint(eval(cmd, globals(), locals()))
                        except:
                            print(traceback.format_exc())
                    wprint("Exiting subshell.")


            ########################
            #  Navigation commands
            #

            elif verb == "show":
                limit = get_int_from_args(args)
                limit = limit if limit else None # get_int_from_args returns False if no value was supplied for the arg
                wprint(resultdb.construct_view(override_options={"limit":limit})[0])
                continue
            elif verb == "showall":
                limit = get_int_from_args(args)
                limit = limit if limit else None # get_int_from_args returns False if no value was supplied for the arg
                wprint(resultdb.construct_view(override_options={"limit":limit,"show_irrelevant":True})[0])
                continue
            elif verb == "dig":
                navigate_ptr = get_ptr_from_id_arg(resultdb, args)
                if navigate_ptr is False:
                    wprint("no [attr] specified, no digging performed.")
                    continue
                if navigate_ptr is None:
                    continue
            elif verb == "up":
                if resultdb.current_pointer.is_base_pointer():
                    wprint("Already at root context.\n")
                    continue
                resultdb.current_pointer.go_up_level()
                # We want to pop out from the categorized_results or grouped_results contexts implicitly.
                if len(resultdb.current_pointer.path) == 1:
                    resultdb.current_pointer.go_up_level()
            elif verb == "top":
                resultdb.navigate_view()
                continue
            elif verb == "dump":
                limit = get_int_from_args(args)
                limit = limit if limit else None # get_int_from_args returns False if no value was supplied for the arg
                wprint(resultdb.results.show_view(limit=limit)[0])
                continue
            elif verb == "exit":
                sys.exit(0)


            ####################################
            #  Data model interaction commands
            #
            
            elif verb == "irrelevant":
                ptr = get_ptr_from_id_arg(resultdb, args)
                if ptr is False:
                    ptr = resultdb.current_pointer
                elif ptr is None:
                    continue
                obj = ptr.give_pointed_object()
                obj.mark_irrelevant()
            elif verb == "relevant":
                ptr = get_ptr_from_id_arg(resultdb, args)
                if ptr is False:
                    ptr = resultdb.current_pointer
                elif ptr is None:
                    continue
                obj = ptr.give_pointed_object()
                obj.mark_relevant()
            elif verb == "fuzzygroup":
                ptr = get_ptr_from_id_arg(resultdb, args)
                if ptr is False:
                    wprint("need to specify an [id], the value of which to group by.")
                    continue
                if ptr is None:
                    wprint(f"couldn't find result object with id {args[1]}.")
                    continue
                obj = ptr.give_pointed_object()
                if isinstance(obj, resultdb.result_class):
                    groupattr = get_attrname_from_attribute_arg(resultdb, args, attr_arg_position=1)
                    if groupattr is False:
                        groupattr = select_attribute(resultdb, "Which attribute of this result would you like to group by? ")
                    elif groupattr is None:
                        continue
                    groupval = obj[groupattr]
                else:
                    wprint("[attr] must point to a specific attribute name of this result.")
                    continue
                group_interactive(resultdb, groupattr, groupval)
            elif verb == "ungroup":
                ptr = get_ptr_from_id_arg(resultdb, args)
                if ptr is False:
                    wprint("need to specify an [id], the value of which to group by.")
                    continue
                if ptr is None:
                    wprint(f"couldn't find result object with id {args[1]}.")
                    continue
                group_ptr = resultdb.current_pointer
                if not any(vertex.value=="grouped_results" for vertex in group_ptr.path):
                    wprint("The current view must be of a result group to use the \"ungroup\" function.")
                    continue
                # We know that ptr points to an object in the current view context, since get_ptr_from_id_arg
                # retrieves the ptr by looking it up in the current view context.
                group = group_ptr.give_pointed_object()
                ungroup_obj_key = ptr.path[-1].value
                del group[ungroup_obj_key]
            elif verb == "find":
                if not len(args):
                    wprint("Need to provide a string to search by.")
                    continue
                filter_str = args[0]
                filtered_attr = get_attrname_from_attribute_arg(resultdb, args, attr_arg_position=1)
                if filtered_attr is False:
                    filtered_attr = select_attribute(resultdb, "Which attribute would you like to filter by? ")
                elif filtered_attr is None:
                    continue
                matches = RelevanceFilteredResultList(result for result in resultdb.results if filter_str in result[filtered_attr])
                wprint(matches.show_view()[0])
            elif verb == "invfind":
                if not len(args):
                    wprint("Need to provide a string to search by.")
                    continue
                filter_str = args[0]
                filtered_attr = get_attrname_from_attribute_arg(resultdb, args, attr_arg_position=1)
                if filtered_attr is False:
                    filtered_attr = select_attribute(resultdb, "Which attribute would you like to filter by? ")
                elif filtered_attr is None:
                    continue
                matches = RelevanceFilteredResultList(result for result in resultdb.results if filter_str not in result[filtered_attr])
                wprint(matches.show_view()[0])
            elif verb == "search":
                if not len(args):
                    wprint("Need to provide a string to search by.")
                    continue
                filter_str = args[0]
                filtered_attr = get_attrname_from_attribute_arg(resultdb, args, attr_arg_position=1)
                if filtered_attr is False:
                    filtered_attr = select_attribute(resultdb, "Which attribute would you like to filter by? ")
                elif filtered_attr is None:
                    continue
                matches = [result for result in resultdb.results if filter_str in result[filtered_attr]]
                resultdb.register_grouped_results(filtered_attr,filter_str,matches)
                wprint(f"Found {len(matches)} results, and saved them as a new result group.")
            elif verb == "invsearch":
                if not len(args):
                    wprint("Need to provide a string to search by.")
                    continue
                filter_str = args[0]
                filtered_attr = get_attrname_from_attribute_arg(resultdb, args, attr_arg_position=1)
                if filtered_attr is False:
                    filtered_attr = select_attribute(resultdb, "Which attribute would you like to filter by? ")
                elif filtered_attr is None:
                    continue
                matches = [result for result in resultdb.results if filter_str not in result[filtered_attr]]
                resultdb.register_grouped_results(filtered_attr,filter_str,matches)
                wprint(f"Found {len(matches)} results, and saved them as a new result group.")
            elif verb == "game":
                ptr = get_ptr_from_id_arg(resultdb, args)
                if ptr is False:
                    obj = resultdb.results
                elif ptr is None:
                    continue
                else:
                    obj = ptr.give_pointed_object()
                play_elimination_game(resultdb, obj)
            elif verb == "filter":
                if not len(args):
                    wprint("Need to provide a string to filter by.")
                    continue
                filter_str = args[0]
                filtered_attr = get_attrname_from_attribute_arg(resultdb, args, attr_arg_position=1)
                if filtered_attr is False:
                    filtered_attr = select_attribute(resultdb, "Which attribute would you like to filter by? ")
                elif filtered_attr is None:
                    continue
                ct = 0
                for result in resultdb.results:
                    if filter_str.lower() in result[filtered_attr].lower():
                        result.mark_irrelevant()
                        ct += 1
                wprint(f"Marked {ct} results as irrelevant using the filter.")
            elif verb == "invfilter":
                if not len(args):
                    wprint("Need to provide a string to filter by.")
                    continue
                filter_str = args[0]
                filtered_attr = get_attrname_from_attribute_arg(resultdb, args, attr_arg_position=1)
                if filtered_attr is False:
                    filtered_attr = select_attribute(resultdb, "Which attribute would you like to filter by? ")
                elif filtered_attr is None:
                    continue
                ct = 0
                for result in resultdb.results:
                    if filter_str.lower() not in result[filtered_attr].lower():
                        result.mark_irrelevant()
                        ct += 1
                wprint(f"Marked {ct} results as irrelevant using the filter.")

            
            
            #####################
            #  Output commands
            #
            
            elif verb == "quiet":
                quieted_attr = get_attrname_from_attribute_arg(resultdb, args)
                if quieted_attr is False:
                    quieted_attr = select_attribute(resultdb, "Which attribute would you like to suppress in output? ")
                elif quieted_attr is None:
                    continue
                ObjectView.global_options["quiet"].append(quieted_attr)
            elif verb == "unquiet":
                attrs = ObjectView.global_options["quiet"]
                if not len(attrs):
                    wprint("No silenced attributes to unquiet.")
                    continue
                quieted_attr = get_attrname_from_attribute_arg(resultdb, args)
                if quieted_attr is False:
                    for i in range(len(attrs)):
                        wprint(f" {i} : {attrs[i]}")
                    wprint()
                    index = int(winput("Which attribute would you like to un-suppress in output? "))
                    quieted_attr = attrs[index]
                elif quieted_attr is None:
                    continue
                elif quieted_attr not in attrs:
                    wprint("That attribute was not silenced, no action taken.")
                    continue
                ObjectView.global_options["quiet"].remove(quieted_attr)
            elif verb == "solo":
                solo_attr = get_attrname_from_attribute_arg(resultdb, args)
                if solo_attr is False:
                    solo_attr = select_attribute(resultdb, "Which attribute would you like to print exclusively in output? ")
                elif solo_attr is None:
                    continue
                ObjectView.global_options["solo"] = solo_attr
            elif verb == "SOLO":
                solo_attr = get_attrname_from_attribute_arg(resultdb, args)
                if solo_attr is False:
                    solo_attr = select_attribute(resultdb, "Which attribute would you like to print exclusively in output? ")
                elif solo_attr is None:
                    continue
                ObjectView.global_options["SOLO"] = solo_attr
            elif verb == "unsolo":
                ObjectView.global_options["solo"] = None
                ObjectView.global_options["SOLO"] = None
            elif verb in ("sort","sortn","rsort","rsortn"):
                if not len(args):
                    wprint("Need a column name to sort by.")
                    continue
                objview = resultdb.current_pointer.give_pointed_object().objectview
                objview["sort_by"] = args[0]
                objview["sort_numeric"] = verb.endswith("n")
                objview["sort_reverse"] = verb.startswith("r")
            elif verb == "history":
                wprint()
                for cmd in command_history:
                    wprint(cmd)
                wprint()
                continue
            elif verb == "width":
                if not len(args):
                    wprint("Need a column name or attribute value to sort by.")
                try:
                    new_width = int(args[0])
                except:
                    wprint(f"Failed to parse {args[0]} as an integer.")
                    continue
                Config.MAX_OUTPUT_WIDTH = new_width
            elif verb == "exportjson":
                if not len(args):
                    wprint("Need a filename to export to.")
                    continue
                try:
                    fname = args[0]
                    ptr = get_ptr_from_id_arg(resultdb, args, id_arg_position=1)
                    if ptr is False:
                        obj = resultdb
                    elif ptr is None:
                        continue
                    else:
                        obj = ptr.give_pointed_object()
                    all_results_json_gen = obj.exportjson()
                    try:
                        if os.path.isfile(fname):
                            answer = winput("WARNING: file already exists. Override? (N/y) ").strip().lower()
                            if not (answer == "y" or answer == "yes"):
                                wprint("Aborting export, no files written.")
                                continue
                            wprint("Overwriting file...")
                        with gzip.GzipFile(fname,"wb") as f:
                            json_encoder = json.JSONEncoder()
                            for result_json in all_results_json_gen:
                                for chunk in json_encoder.iterencode(result_json):
                                    f.write(chunk.encode('utf-8'))
                        wprint(f"Export success, compressed JSON output written to {fname}.")
                    except PermissionError:
                        wprint("Failed to open the specified file, maybe try an absolute path? (FYI, quotes are supported.)")
                    continue
                except Exception as e:
                    wprint(f"Exception encountered while exporting: {e}")
                    continue
            elif verb == "export":
                if not len(args):
                    wprint("Need a filename to export to.")
                    continue
                try:
                    fname = args[0]
                    ptr = get_ptr_from_id_arg(resultdb, args, id_arg_position=1)
                    if ptr is False:
                        # no specific object (result, or row, etc) was specified - export everything
                        obj = resultdb
                    elif ptr is None:
                        wprint("Couldn't find the specified result, no action taken.")
                        continue
                    else:
                        obj = ptr.give_pointed_object()
                    resultgen = obj.all_result_objects()
                    try:
                        if os.path.isfile(fname):
                            answer = winput("WARNING: file already exists. Override? (N/y) ").strip().lower()
                            if not (answer == "y" or answer == "yes"):
                                wprint("Aborting export, no files written.")
                                continue
                            wprint("Overwriting file...")
                        with gzip.GzipFile(fname,"wb") as f:
                            # pickle-dump the results one by one, allowing them to be loaded one-by-one when loading back.
                            # https://stackoverflow.com/questions/20716812/saving-and-loading-multiple-objects-in-pickle-file
                            for result in resultgen:
                                pickle.dump(result, f)
                        wprint(f"Export success, compressed serialized output written to {fname}.")
                    except PermissionError:
                        wprint("Failed to open the specified file, maybe try an absolute path? (FYI, quotes are supported.)")
                    continue
                except Exception as e:
                    wprint(f"Exception encountered while exporting: {e}")
                    continue
            else:
                wprint("Unrecognized command.\n")
                continue
            
            if navigate_ptr:
                current_view = resultdb.navigate_view(navigate_ptr)
            else:
                current_view = resultdb.update_current_view()
        except KeyboardInterrupt:
            sys.exit(0)
        except SystemExit:
            raise
        except:
            print(traceback.format_exc())

        
HOME_DIRECTORY = str(Path.home())
WHITTLER_DIRECTORY = HOME_DIRECTORY+"/.whittler"
if not os.path.isdir(WHITTLER_DIRECTORY):
    os.mkdir(WHITTLER_DIRECTORY,mode=0o770)

result_classes = {}

def load_module(fname, in_default_module_directory):
    if in_default_module_directory:
        module = importlib.import_module(f"Whittler.modules.{fname[:fname.index('.')]}")
    else:
        modulepath = os.path.abspath(fname)
        folder, fname = os.path.split(modulepath)
        if not fname.endswith(".py"):
            raise Exception("Module names must end in .py")
        loader = importlib.machinery.SourceFileLoader(fname[:fname.index('.')], modulepath)
        module = loader.load_module()
    result_class = None
    for clsname in dir(module):
        if clsname.startswith("__"):
            continue
        cls = getattr(module, clsname)
        if isinstance(cls, type) and issubclass(cls, Result) and not cls is Result:
            result_classes[cls.FRIENDLY_NAME] = cls
            if not result_class is None:
                raise Exception(f"Multiple Result subclasses defined in the module loaded from {module.__file__}")
            result_class = cls
    if not result_class:
        raise Exception(f"Could not find any Result subclass in the module loaded from {module.__file__}")
    return result_class

# load default result classes
for fname in filter(lambda s: not s.startswith("_") , os.listdir(WHITTLER_DIRNAME+"/modules")):
    load_module(fname, True)

# This class is a hack for the choices= option of the --config argparse argument definition. I want the help
# message to enumerate the default modules in the Whittler/modules folder, but I also want the user to be able
# to specify a custom file to try to use as a module.
class ConfigChoices(list):
    def __contains__(self,value):
        return True

config_choices = ConfigChoices(list(result_classes.keys()))


# to be populated if the --script or --scriptfile flags are specified
cached_commands = []

def main():
    global cached_commands
    # This is a patch for running the script with "python -m Whittler" - without this modification the help messages
    # think the script is named "__main__.py".
    old_argv0 = sys.argv[0]
    sys.argv[0] = old_argv0.replace("__main__","Whittler")

    parser = argparse.ArgumentParser(description="An interactive script to whittle down large datasets")

    # Reset the script name back to the old value, just in case.
    sys.argv[0] = old_argv0

    # Required args
    bargs = parser.add_argument_group("basic arguments")
    bargs.add_argument('--config',
                    help='the module to use to parse the specified tool output files.',
                    type=str, nargs=1, choices=config_choices, default=None, required=True)

    # Data ingestion args
    diargs = parser.add_argument_group("data ingestion arguments")
    #diargs = diargs.add_mutually_exclusive_group(required=True)
    diargs.add_argument('--file',
                        help='the tool output file to be parsed',
                        type=str, nargs='+', default='')
    diargs.add_argument('--dir',
                        help='the directory containing tool output files to be parsed',
                        type=str, nargs='+', default='')
    diargs.add_argument('--import_whittler_output',
                        help='consume and continue working with one or more files that were outputted by Whittler\'s "export" command',
                        type=str, nargs='+', default=None, metavar="FILE_OR_DIR")
    diargs.add_argument('--memory_compression',
                        help='enable in-memory compression for working with very large datasets',
                        action='store_true')
    diargs.add_argument('--multiprocessing',
                        help='enable multiprocessing for working with very large datasets (in development, currently only '+\
                             'accelerates imports from directories with many files)',
                        action='store_true')

    # Output control args
    ocargs = parser.add_argument_group("output control arguments")
    ocargs.add_argument('--log_output',
                        help='a file to which all output in this session will be logged (default: a new file in the '+\
                            '.whittler folder in your home directory)',
                        type=str, nargs="?", default=None, metavar="FILENAME",
                        const=WHITTLER_DIRECTORY+'/{date:%Y-%m-%d_%H-%M-%S}_log.txt'.format( date=datetime.datetime.now() ) )
    ocargs.add_argument('--log_command_history',
                        help='a file in which to record the command history of this session, in a format that can be '+\
                            'imported and re-run by the --scriptfile flag (default: a new file in the .whittler folder in '+\
                            'your home directory)',
                        type=str, nargs="?", default=None, metavar="FILENAME",
                        const=WHITTLER_DIRECTORY+'/{date:%Y-%m-%d_%H-%M-%S}_command_log.txt'.format( date=datetime.datetime.now() ))

    # Scripting arguments
    sargs = parser.add_argument_group("scripting arguments")
    sargs.add_argument('--script',
                    help='run a script specified with a string on the command line, with each command separated by semicolons '+\
                            '(backslash-escape for a literal semicolon)',
                    type=str, nargs=1, default=None, metavar="SCRIPT_STRING")
    sargs.add_argument('--scriptfile',
                    help='run a script provided in a file, with one command per line',
                    type=str, nargs=1, default=None, metavar="SCRIPT_FILE")

    args = parser.parse_args()

    if args.memory_compression:
        Config.MEMORY_COMPRESSION = True
        Config.MemoryCompressor.COMPRESSION_ENABLED = True
    
    if args.multiprocessing:
        Config.MULTIPROCESSING = True
        Config.MemoryCompressor.MULTIPROCESSING = True

    if not args.log_output is None:
        logdir = args.log_output[0] if isinstance(args.log_output,list) else args.log_output
        try:
            set_global_redirect_file(open(logdir,"w+", encoding="utf-8"))
        except PermissionError:
            wprint("Lacking permissions to write to the specified all-output log file.")
            sys.exit(1)
        get_global_redirect_file().write(" ".join(sys.argv)+"\n\n")
    if not args.log_command_history is None:
        logcmddir = args.log_command_history[0] if isinstance(args.log_command_history,list) else args.log_command_history
        try:
            set_command_redirect_file(open(logcmddir,"w+", encoding="utf-8"))
        except PermissionError:
            wprint("Lacking permissions to write to the specified command history log file.")
            sys.exit(1)
    if not args.config or len(args.config) != 1:
        parser.print_help()
        sys.exit(1)
    try:
        config = args.config[0]
        if config in result_classes:
            resultdb = ResultDatabase(result_classes[config])
        else:
            resultdb = ResultDatabase(load_module(config, False))
        if not args.dir and not args.file and not args.import_whittler_output:
            parser.print_help()
            sys.exit(1)
        wprint("\nWelcome to the Whittler shell. Type \"help\" for a list of commands.\n")
        if args.script and args.scriptfile:
            wprint("Error: the --script and --scriptfile flags cannot both be specified at once.\n")
            parser.print_help()
            sys.exit(1)
        if args.script or args.scriptfile:
            if args.script:
                wprint(f"Using script from command line.")
                cached_commands = parse_user_input(args.script[0])
            else:
                wprint(f"Importing script from file {args.scriptfile[0]} .")
                with open(args.scriptfile[0],"r") as f:
                    cached_commands = parse_user_input(" ; ".join(f.readlines()))
        hash_cache = set()
        if args.dir:
            wprint("Parsing files from provided directories...")
            wprint()
            for d in args.dir:
                resultdb.parse_from_directory(d, hash_cache=hash_cache, multiprocessing_import=args.multiprocessing)
        if args.file:
            wprint("Parsing provided files...")
            wprint()
            for fname in args.file:
                resultdb.parse_from_file(fname, hash_cache=hash_cache)
        if args.import_whittler_output:
            wprint("Importing provided files...")
            wprint()
            for import_target in args.import_whittler_output:
                if os.path.isdir(import_target):
                    import_target = [import_target+"/"+fname for fname in os.listdir(import_target)]
                else:
                    import_target = [import_target]
                for fname in import_target:
                    resultdb.parse_from_export(fname, hash_cache=hash_cache)
        
        if Config.MEMORY_COMPRESSION and Config.MemoryCompressor.training_mode:
            Config.MemoryCompressor.disable_training_mode()

        wprint("Done.\n")
        main_loop(resultdb)
    except:
        raise
    finally:
        print()
        if not command_redirect_file is None:
            fname = command_redirect_file.name.replace('\\','/')
            wprint(f"Saving logged command history to {fname} ...")
            command_redirect_file.close()
        if not redirect_file is None:
            fname = redirect_file.name.replace('\\','/')
            wprint(f"Saving pipe output to {fname} ...")
            redirect_file.close()
        goodbye_said = False
        if not global_redirect_file is None:
            fname = global_redirect_file.name.replace('\\','/')
            wprint(f"Saving logged output to {fname} ...")
            wprint("\nGoodbye.\n")
            goodbye_said = True
            global_redirect_file.close()
        if not goodbye_said:
            print("\nGoodbye.\n")
        

if __name__ == "__main__":
    main()
